c Uso: 
c - Eseguire: php hcVip.php [-smart|-straightforward] -instance <file istanza>
c
c Esecuzione con zChaff su istanza di test e successiva traduzione del modello trovato in forma intellegibile
c php hcVip.php -instance hcVip-sampleInstance.txt > cnf.txt ; zchaff cnf.txt | php zchaff2hcVip.php cnf.txt
c
c Hamiltonian Cycle (with important destinations) to SAT compiler (the most straightforward and inefficient encoding, just a simple exercise!)
c Author: Toni Mancini - Sapienza University of Rome
c Date: September 18, 2008
c
<?php

	$DEBUG = false;  // <-- Set this variable to true to generate a more intelligible (non-DIMACS) 
					 //     CNF (variables with symbolic names and few comments)


	// Variables
	$DEST = NULL;
	$VIP = NULL;
	$HOME = NULL;
	$BUS = array();  // 2D matrix: from/to

	$SMART = true;  // smarting encoding by default (constraint 8): to disable it, use -straightforward

    $fileName = null;
	for ($i=1; $i<count($argv); $i++) {
		if ($argv[$i] === "-smart") {
			$SMART = true;
		} else if ($argv[$i] === "-straightforward") {
			$SMART = false;
		} else if ($argv[$i] === "-instance") {
			$fileName = $argv[++$i];
		}
	}

	if ($fileName == "") {
		echo("c Syntax: php ".$argv[0]." [-smart|-straightforward] -instance <instance>\n");
		exit;
	};
	
	readInstanceFile($fileName);


	// Check that everything has been given
	$DEST != NULL or die("Number of destinations missing\n");
	$VIP != NULL or die("List of VIPs missing\n");
	$HOME != NULL or die("Home missing\n");
	
	$MAXTIME = $DEST;   // Path is long $DEST steps, plus a last step to go back home.
	
	
	// Everything is ok (`bus' is not mandatory, it may be empty! Of course in this case, the istance is unsatisfiable)
	
	
	$CLAUSES = array();
	$NCLAUSES = 0;
	$NVARS = 0;
	$dictionary = NULL; 
	/* will be a 3D array: 
		dictionary['VISIT'][dest][time] =  dimacsVar for visit(dest,time)		
	*/
	
?>
c
c Problem description (a variation of the Hamiltonian Cycle problem):
c
c Given:   
c  - $DEST destinations (numbered 1 to $DEST),    
c  - a distinguished destination, `home' (numbered $HOME),    
c  - a subset of the destinations considered "important" (VIP)   
c  - a set of bus lines that connect (ordered) pairs of destinations,   
c     
c  we want to find a cycle that starts an terminates at home, such that:   
c  - All destinations are visited exactly once   
c  - It is possible to reach any destination from the previous one with a single   
c    bus trip   
c  - All VIP destinations are visited in the first half of the path   
c
<?php
	echo("c Number of destinations: $DEST\n");
	echo("c VIP destinations: [" . join(", ", $VIP) ."]\n");
	echo("c Home: $HOME\n");
	echo("c Bus lines: omitted\n");	

	echo("c\nc " . ($SMART?"smart":"straightforward") . " encoding\n");
	//print_r($BUS);


	/* Variables:
        We define a propositional variable `visit_d_t' for each d in 1..$DEST and each 
        t in 1..$MAXTIME (time)

        Variable visit_d_t is true iff we have to visit destination `d' at time `t'

	*/



	if ($DEBUG) addComment("c (1) Each destination is visited at least once");
	/* 
		AND_{d:1..$DEST} OR_{t:1..$MAXTIME}  visit_d_t	
	*/
	
	for($d=1; $d<=$DEST; $d++) {
		$clause = "";
		for($t=1; $t<=$MAXTIME; $t++) {
			$clause .= visit2dimacs($d,$t) . " ";
		}
		addClause($clause);
	}
	


	if ($DEBUG) addComment("c (2) Each destination is visited at most once");
	/* 	
		 Note that home is visited a second time at the $MAXTIME+1 step, but this is outside our path that models trip up to the last destination (this requirement will be enforced by additional constraint 5., cf. below)
		
		AND_{d:1..$DEST} AND_{t1, t2:1..$MAXTIME and t1!=t2}  -visit_d_t1 || -visit_d_t2	
	*/
	for($d=1; $d<=$DEST; $d++) {
		for($t1=1; $t1<=$MAXTIME; $t1++) {	
			for($t2=$t1+1; $t2<=$MAXTIME; $t2++) {	  // Avoid generating "duplicate" clauses
				addClause( "-" . visit2dimacs($d,$t1) . " -".visit2dimacs($d,$t2));
			}
		}
	}


	if ($DEBUG) addComment("c (3) At each time step at least one destination is visited");
	/* 
		AND_{t:1..$MAXTIME} OR_{d:1..$DEST}  visit_d_t	
	*/

	for($t=1; $t<=$MAXTIME; $t++) {	
		$clause = "";
			for($d=1; $d<=$DEST; $d++) {
			$clause .= visit2dimacs($d,$t) . " ";
		}
		addClause($clause);
	}
	


	if ($DEBUG) addComment("c (4) At each time step at most one destination is visited");
	/* 	
		AND_{t:1..$MAXTIME} AND_{d1, d2:1..$DEST and d1!=d2}  -visit_d1_t || -visit_d2_t	
	*/
	for($t=1; $t<=$MAXTIME; $t++) {
		for($d1=1; $d1<=$DEST; $d1++) {	
			for($d2=$d1+1; $d2<=$DEST; $d2++) {	  // Avoid generating "duplicate" clauses
				addClause( "-" . visit2dimacs($d1,$t) . " -".visit2dimacs($d2,$t));
			}
		}
	}

	

	if ($DEBUG) addComment("c (5) At each step, the next destination must be reachable by a single bus trip");
	/* 
		AND_{d1,d2:1..$DEST s.t. not bus(d1,d2)} 
      		AND_{t:0..$MAXTIME-1} -visit_d1_t || visit_d2_(t+1)
	*/
	for($d1=1; $d1<=$DEST; $d1++) {
		for($d2=1; $d2<=$DEST; $d2++) {
			if ($d1 == $d2) continue;
			if ($BUS[$d1][$d2] == 0) {			
				if ($DEBUG) addComment("c ...There is no bus from $d1 to $d2:");		
				for($t=1; $t<=$MAXTIME-1; $t++) {	
					addClause( "-" . visit2dimacs($d1,$t) . " -".visit2dimacs($d2,$t+1));
				}
			}
		}
	}
	
	

	if ($DEBUG) addComment("c (6) Path starts from home (home in 1..$DEST)");
	/* visit_home_1
	*/
	addClause( visit2dimacs($HOME,1) );



	if ($DEBUG) addComment("c (7) Home must be reachable from the last destination by a single bus trip");
	/* 
		AND_{d:1..$DEST s.t. not bus(d,home)}  -visit_d_($MAXTIME)	
	*/
	for($d=1; $d<=$DEST; $d++) {
		if ($BUS[$d][$HOME] == 0) {
			if ($DEBUG) addComment("c ...There is no bus from $d to home");
			addClause( "-" . visit2dimacs($d,$MAXTIME) );
		}
	}

	if ($DEBUG) addComment("c (8) VIP customers are visited in the first half of the path");
	/* 
		The most intuitive formalization would be:
			AND_{d:V} OR_{t=1..floor(($MAXTIME)/2)} visit_d_t.
		
		It can be observed how such encoding would lead to long clauses (inefficient to solve)
		
		An alternative formalization is the following one:
			"any VIP destination must NOT be in the second half of the path
				AND_{d:V} AND_{t=floor(($MAXTIME)/2)+1 .. $MAXTIME} - visit_d_t

		that will produce only UNARY clauses.	
	*/
	
	if (!$SMART) {
		// Most intuitive formalization
		foreach($VIP as $d) {
			$clause = "";
			for($t=1; $t <= $MAXTIME/2; $t++) {
				$clause .= visit2dimacs($d,$t) . " ";
			}
			addClause( $clause );
		}
	} else {	
		// Smarter formalization
		foreach($VIP as $d) {
			for($t=($MAXTIME/2)+1; $t<=$MAXTIME; $t++) {
				addClause( "-" . visit2dimacs($d,$t) );
			}
		}
	}	


	// Print out dictionary
	echoDictionary();
	

	
	// Print out header line "p cnf..."
	isset($NVARS) or die("maxVar is not set");
	echo("p cnf $NVARS $NCLAUSES\n");
	
	
	// Print out all clauses
	echoClauses();	

	//print_r($dictionary); // debug only
	
	


	function addClause($clause) {
		global $CLAUSES,$NCLAUSES;
		$clause = trim($clause);
		if ($clause == "") return;
		$clause .= " 0";
		$CLAUSES[] = $clause; // Append to array
		$NCLAUSES++;
	}

	function addComment($msg) {
		global $CLAUSES;
		$CLAUSES[] = $msg;
		// No $NCLAUSES increment here!
	}


	// 	dictionary['VISIT'][dest][time]
	function visit2dimacs($d,$t) {
		global $DEBUG, $NVARS, $dictionary, $DEST, $VIP, $HOME, $BUS;
		
		if ($d < 1 || $d > $DEST) {
			echoClauses();
			die("Trying to retrieve a 'VISIT' variable for destination $d: VISIT($d,$t)");
		}
		if ($t < 1 || $t > $DEST) {
			echoClauses();
			die("Trying to retrieve a 'VISIT' variable for time $t: VISIT($d,$t)");
		}
		$result = "VISIT(d=$d,t=$t)";		
		
		$v = $dictionary['VISIT'][$d][$t];
		if ($v !== NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['VISIT'][$d][$t] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}
	

	
	function echoClauses() {
		global $CLAUSES;
		foreach($CLAUSES as $c) {
			echo($c ."\n");
		}
	}	
	
		
	
	function echoDictionary() {
		global $DEBUG, $NVARS, $dictionary, $DEST, $VIP, $HOME, $BUS;
		echo ("c\nc START DICTIONARY\n");
		/*if ($DEBUG) {
			echo("c omitted -- DEBUG mode\n");
		}
		else {*/		
			if (isset($dictionary['VISIT'])) {			
				// 	dictionary['VISIT'][dest][time] =  dimacsVar
				foreach($dictionary['VISIT'] as $d => $entryD) {
					foreach($entryD as $t => $dimacsVar) {
						echo("c visit($d,$t) = $dimacsVar\n");
					}
				}
			}		
		echo("c END DICTIONARY\nc\nc\n");		
	}	
	
	

	

	
	
	function readInstanceFile($fileName) {
		global $DEBUG, $NVARS, $dictionary, $DEST, $VIP, $HOME, $BUS;

		$COMMENTS = array("#", "//");		
		$file = file($fileName) or die("Could not open instance file $fileName\n");
		foreach($file as $line) {
			//echo("c Reading line: $line\n");
			$commentSign = NULL;
			foreach($COMMENTS as $c) {
				 $pos = strpos($line, $c);
				 if ($pos === FALSE) continue;
				 if ($commentSign === NULL || $commentSign > $pos) $commentSign = $pos;
			}
			if ($commentSign !== NULL) {
				$line = substr($line, 0, $commentSign);
				//echo("   c Removing comments (from col. $commentSign onwards): $line\n");		
			}
			$line = trim($line);
			if ($line == "") continue;
	
			list($what,$content) = split("=",$line);
			$what = trim($what);
			$content = trim($content);
	
			//echo("   c What: $what -- Content: $content\n");		
			
			if ($what == "destinations" && $DEST===NULL) {
				$DEST = $content;
				if (!is_numeric($DEST) || $DEST < 1) 
					die("$DEST is not a good value for `destinations'\n");

				for ($from=1; $from<=$DEST; $from++) {
					for($to=1; $to<=$DEST; $to++) {
						$BUS[$from][$to] = 0;
					}
				}
			}
			else if ($what == "vip" && $VIP===NULL) {
				$VIP = explode(",",$content);
				for($i=0; $i<count($VIP); $i++) {
					$VIP[$i] = trim($VIP[$i]); // Remove spaces near commas
					if (!is_numeric($VIP[$i])) die("Error: `".$VIP[$i]."' (in array VIP) is not a number\n");
					if ($VIP[$i] < 1 || $VIP[$i] > $DEST) die("Error: `".$VIP[$i]."' (in array VIP) is not in range 1..$DEST\n");
				}					
			}
			else if ($what == "home" && $HOME===NULL) {
				$HOME = $content;
				if (!is_numeric($HOME) || $HOME < 1 || $HOME > $DEST) 
					die("Error: `".$HOME."' is not a valid value for home\n");			
			}
			else if ($what == "bus") {
				list($from, $to) = explode("->", $content);
				$from = trim($from);
				$to = trim($to);
				if (!is_numeric($from)) die("Error: `".$from."' (bus line ".$from."->".$to.") is not a number\n");
				if (!is_numeric($to)) die("Error: `".$to."' (bus line ".$from."->".$to.") is not a number\n");
				
				$BUS[$from][$to] = 1;  // Add bus line
			}
		}
	}	
?>
